/* $Id: nand-gpbc.c,v 1.1.1.1 2016/01/29 02:41:41 txbsd Exp $
 * nand-gpbc.c: MTD NAND flash driver for gpbc
 *
 * Copyright (C) 2010-2011 Panasonic Corporation
 * All Rights Reserved.
 *
 * SPDX-License-Identifier:	GPL-2.0
 */


#include <common.h>
#include <malloc.h>
#include <nand.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/compat.h>
#include <linux/err.h>
#include <asm/errno.h>
#include <asm/io.h>

#include <linux/stddef.h>

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
#include <asm/cache.h>
#include <asm/arch/cacheflush.h>
#endif

#include "nand-gpbc.h"
#include "nand-gpbc-reg.h"

#if !((0 < NAND_GPBC_FLASH_BANKS) && (NAND_GPBC_FLASH_BANKS <= NAND_GPBC_MAX_FLASH_BANKS))
#error "Error: Invalid NAND_GPBC_FLASH_BANKS."
#endif

#if 1 /* for compatibility with Linux */
#define NAND_MAX_CHIPS		CONFIG_SYS_NAND_MAX_CHIPS
#define PURGE_CACHE_D_PURGE	PURGE_CACHE_D_PURGE_INV
#define pr_cont(args...)	MTDDEBUG(MTD_DEBUG_LEVEL0, args)
#endif

#define GPBC_TBD

static uint8_t nand_gpbc_bbt_pattern[] = NAND_GPBC_UNIT_BBT_PATTERN_MASTER;
static struct nand_bbt_descr nand_gpbc_BbtMasterDesc = {
	.options	= NAND_GPBC_BBT_OPTION,
	.offs		= NAND_GPBC_BBT_OFFS,
	.len		= NAND_GPBC_BBT_LEN,
	.veroffs	= NAND_GPBC_BBT_VEROFFS,
	.maxblocks	= NAND_GPBC_UNIT_BBT_MAXBLOCKS,
	.pattern	= nand_gpbc_bbt_pattern
};
static uint8_t nand_gpbc_mirror_pattern[] = NAND_GPBC_UNIT_BBT_PATTERN_MIRROR;
static struct nand_bbt_descr nand_gpbc_BbtMirrorDesc = {
	.options	= NAND_GPBC_BBT_OPTION,
	.offs		= NAND_GPBC_BBT_OFFS,
	.len		= NAND_GPBC_BBT_LEN,
	.veroffs	= NAND_GPBC_BBT_VEROFFS,
	.maxblocks	= NAND_GPBC_UNIT_BBT_MAXBLOCKS,
	.pattern	= nand_gpbc_mirror_pattern
};

static struct nand_ecclayout nand_gpbc_EccLayout = {
	.eccbytes	= 0,
	.eccpos		= NAND_GPBC_UNIT_OOB_ECCPOS,
	.oobfree	= NAND_GPBC_UNIT_OOB_OOBFREE
};

#ifdef CONFIG_MTD_NAND_MN2WS_BB
static uint8_t nand_gpbc_bbt_flag_ff_pattern[] = NAND_GPBC_BBT_FLAG_FF_PATTERN;
static struct nand_bbt_descr nand_gpbc_BbtFlagDesc = {
	.options	= (NAND_GPBC_BBT_OPTION & (~NAND_BBT_NO_OOB)),
	.offs		= NAND_GPBC_BBT_FLAG_OFFS,
	.len		= NAND_GPBC_BBT_FLAG_LEN,
	.pattern	= nand_gpbc_bbt_flag_ff_pattern
};
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

static uint8_t nand_gpbc_scan_ff_pattern[] = NAND_GPBC_BBTSCAN_FF_PATTERN;
static struct nand_bbt_descr nand_gpbc_BbPatternDesc = {
	.options	= (NAND_GPBC_BBT_OPTION & (~NAND_BBT_NO_OOB)),
	.offs		= NAND_GPBC_BBTSCAN_OFFS,
	.len		= NAND_GPBC_BBTSCAN_LEN,
	.pattern	= nand_gpbc_scan_ff_pattern
};

#ifdef CONFIG_MTD_NAND_MN2WS_BB
static struct nand_bbm	nand_gpbc_ABbm[NAND_MAX_CHIPS];
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

/* this macro allows us to convert from an MTD structure to our own
 * device context (gpbc) structure.
 */
//#define mtd_to_gpbc(m) container_of(m, struct nand_gpbc_info, mtd)
#define mtd_to_gpbc(m) ((struct nand_gpbc_info *)((struct nand_chip *)m->priv)->priv)

static void nand_read_id(struct nand_gpbc_info *gpbc, u8 *nand_id, int num);

//----------------------------
//------ layer 1 func --------
//----------------------------
static void nand_gpbc_index32(struct nand_gpbc_info *gpbc, u32 cmd, u32 data)
{
	*(volatile u32 *)gpbc->memBase = cmd;
	*(volatile u32 *)(gpbc->memBase + 0x10) = data;
}

static void nand_gpbc_write32(u32 data, void *addr)
{
	*(volatile u32 *)addr = data;
}

static u32 nand_gpbc_read32(void *addr)
{
	u32 data;

	data = *(volatile u32 *)addr;
	return data;
}

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_RESET
static void nand_gpbc_mask_write32(u32 data, u32 mask, void *addr)
{
	u32 value;

	value = (nand_gpbc_read32(addr) & ~(mask)) | (data & mask);
	nand_gpbc_write32(value, addr);
}
#endif

static void nand_gpbc_clearIntStatus(struct nand_gpbc_info *gpbc, int bank)
{
	nand_gpbc_write32(NAND_GPBC_NI__MASK, gpbc->regBase + gpbc->nist[bank]);
}

#ifndef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
static u32 nand_gpbc_pollIrq(struct nand_gpbc_info *gpbc, int bank, u32 mask)
{
	u32		result;
	unsigned long	timeo;

	//sync
	(void)nand_gpbc_read32(gpbc->regBase + gpbc->nie[bank]);

	reset_timer();
	do {
		result = nand_gpbc_read32(gpbc->regBase + gpbc->nist[bank]);
		if (result & mask) {
			return result;
		}
		timeo = get_timer(0);
	} while (((u64)NAND_GPBC_CMD_TIMEOUT * CONFIG_SYS_HZ + 999) / 1000 > timeo);
	result = nand_gpbc_read32(gpbc->regBase + gpbc->nist[bank]);
	if (!(result & mask)) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	}

	return result;
}
#endif

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
static irqreturn_t nand_gpbc_irq(int this_irq, void *dev_id)
{
	struct nand_gpbc_info *gpbc = (struct nand_gpbc_info *)dev_id;

	nand_gpbc_write32(0, gpbc->regBase + gpbc->nie[0]);	//disable irq
	nand_gpbc_write32(0, gpbc->regBase + gpbc->nie[1]);	//disable irq
	(void)nand_gpbc_read32(gpbc->regBase + gpbc->nie[1]);	//sync
	gpbc->event = 1;
	wake_up(&gpbc->queue);

	return IRQ_HANDLED;
}

static u32 nand_gpbc_waitIrq(struct nand_gpbc_info *gpbc, int bank, u32 mask)
{
	long	status;
	u32		result;

	gpbc->event = 0;
	nand_gpbc_write32(mask, gpbc->regBase + gpbc->nie[bank]);	//sync & enable irq
	status = wait_event_timeout(gpbc->queue, (gpbc->event == 1)
				    , msecs_to_jiffies(NAND_GPBC_IRQ_TIMEOUT));
	nand_gpbc_write32(0, gpbc->regBase + gpbc->nie[bank]);	//disable irq : abnormal condition
	result = nand_gpbc_read32(gpbc->regBase + gpbc->nist[bank]);
	if (!status && (gpbc->event != 1)) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		panic( "%s:%d\n", __FILE__, __LINE__ );
	}

	return result;
}
#endif

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
static void nand_gpbc_set_cmd_desc_read_page(struct nand_gpbc_cmd_desc_t *cmd_desc,
					     u32 flash_page, u16 cmd_flags, u64 mem_ptr)
{
	cmd_desc->next_ptr = 0;
	cmd_desc->flash_page = flash_page;
	cmd_desc->reserved0 = 0;
	cmd_desc->cmd_type = 0x2001;	//read-ahead
	cmd_desc->cmd_flags = cmd_flags;
	cmd_desc->sync_arg = 0;
	cmd_desc->mem_ptr = mem_ptr;
	cmd_desc->status = 0;
	cmd_desc->reserved1 = 0;
	cmd_desc->reserved2 = 0;
	cmd_desc->sync_flag_ptr = 0;
	cmd_desc->mem_copy_addr = 0;
	cmd_desc->meta_data_addr = 0;
}

static u32 nand_gpbc_poll_cmd_dma(struct nand_gpbc_info *gpbc)
{
	volatile struct nand_gpbc_cmd_desc_t *cmd_desc = gpbc->cmd_desc;
	while (1) {
		if (cmd_desc->status & NAND_GPBC_CDESC_STAT_CMP) {
			break;
		}
	}
	return (u32)cmd_desc->status;
}
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */

#ifndef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
static u32 nand_gpbc_checkIrq(struct nand_gpbc_info *gpbc, int bank)
{
	u32		result;

	result = nand_gpbc_read32(gpbc->regBase + gpbc->nist[bank]);

	return result;
}
#endif

static int nand_gpbc_update_correct_info(struct mtd_info *mtd, int page,
					 int num)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	struct nand_gpbc_correct_info *info = &gpbc->correct_info;

	if (num < info->num) {
		return -1;
	}

	info->page = page;
	info->num = num;

	return 0;
}

static int nand_gpbc_get_correct_info(struct mtd_info *mtd,
				      struct nand_correct_info_t *correct_info)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	struct nand_gpbc_correct_info *info = &gpbc->correct_info;

	if (info->page == NAND_CORRECT_NO_PAGE) {
		return -ENOENT;
	}
	correct_info->page = info->page;
	return info->num;
}

static void nand_gpbc_clear_correct_info(struct mtd_info *mtd)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	struct nand_gpbc_correct_info *info = &gpbc->correct_info;

	info->page = NAND_CORRECT_NO_PAGE;
	info->num = 0;
}

static int nand_gpbc_get_max_bitflips(struct nand_gpbc_info *gpbc, int bank)
{
	int max_bitflips;

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
	max_bitflips = (gpbc->cmd_desc->status & NAND_GPBC_CDESC_STAT_MAX_ERR) >>
		NAND_GPBC_CDESC_STAT_MAX_ERR_SHIFT;
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */
	max_bitflips = (nand_gpbc_read32(gpbc->regBase + NAND_GPBC_NECCINF) >> (8 * bank)) & 0x7f;
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */

	return max_bitflips;
}


//----------------------------
//------ layer 2 func --------
//----------------------------
static void nand_gpbc_reset_buf(struct nand_gpbc_info *gpbc)
{
	gpbc->buf.head = gpbc->buf.tail = 0;
}

static void nand_gpbc_write_byte_to_buf(struct nand_gpbc_info *gpbc, uint8_t byte)
{
	BUG_ON(gpbc->buf.tail >= sizeof(gpbc->buf.buf));
	gpbc->buf.buf[gpbc->buf.tail++] = byte;
}

static void nand_gpbc_read_status(struct nand_gpbc_info *gpbc)
{
	uint32_t cmd = 0x0;

	nand_gpbc_reset_buf(gpbc);

	cmd = nand_gpbc_read32(gpbc->regBase + NAND_GPBC_NWPD);
	if (cmd & NAND_GPBC_NWPD__WPD) {
		nand_gpbc_write_byte_to_buf(gpbc, NAND_STATUS_WP);
	} else {
		nand_gpbc_write_byte_to_buf(gpbc, 0);
	}
}

static u32 nand_gpbc_resetBank(struct nand_gpbc_info *gpbc, int bank)
{
	u32		irqMask;
	u32		result;

	nand_gpbc_clearIntStatus(gpbc, bank);
	nand_gpbc_write32((NAND_GPBC_NDRST__B0RST<<bank), gpbc->regBase + NAND_GPBC_NDRST);
	irqMask = NAND_GPBC_NI__RSTCMP | NAND_GPBC_NI__ERR;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irqMask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irqMask);
#endif
	if (NAND_GPBC_NI__ERR & result) {
		return result;
	}

	return 0;
}

static int nand_gpbc_isErasedPageData(struct nand_gpbc_info *gpbc, u8 *buf)
{
	struct nand_chip *chip = &gpbc->chip;
	u32 base, cnt, sect, calc;
	u32 correct_type = nand_gpbc_read32(gpbc->regBase + NAND_GPBC_NECCC);

	calc = 0xffffffff;

	for (sect = 0; sect < chip->ecc.steps; sect++) {
		base = chip->ecc.size * sect;
		cnt = 0;

		if (correct_type == 8) {
			//It is tested in only 8bit Ecc.
			u32 corrected_bit_offset = 0x2c0;
			for (; cnt < corrected_bit_offset;
			     cnt += sizeof(u32)) {
				calc &= *(u32 *)(buf + base + cnt);
			}
			calc &= (*(u32 *)(buf + base + cnt) ^ 0x00000200);
			cnt += sizeof(u32);
		}

		for (; cnt < chip->ecc.size; cnt += sizeof(u32)) {
			calc &= *(u32 *)(buf + base + cnt);
		}
	}

	if (calc != 0xffffffff) {
		return 0;
	}

	return 1;
}

static int nand_gpbc_isErasedPageEcc(struct nand_gpbc_info *gpbc, int page, u8 *buf)
{
	struct nand_chip *chip = &gpbc->chip;
	u32 irq_mask;
	u32 result;
	u32 sect, sect_base, ecc_base, start, end, cnt;
	u8 calc;
	int bank = gpbc->flash_bank;
	u32 bank_sel = bank << NAND_GPBC_BANK_SEL;

	nand_gpbc_clearIntStatus(gpbc, bank);
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank_sel | NAND_GPBC_MAP11_CMD, NAND_CMD_READ0);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank_sel | NAND_GPBC_MAP11_ADR, 0);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank_sel | NAND_GPBC_MAP11_ADR, 0);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank_sel | NAND_GPBC_MAP11_ADR, page & 0xff);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank_sel | NAND_GPBC_MAP11_ADR, (page >> 8) & 0xff);
	if ((CONFIG_MTD_NAND_MN2WS_CHIP_CA_BYTES == 2) &&
	    (CONFIG_MTD_NAND_MN2WS_CHIP_PA_BYTES == 3)) {
		nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank_sel | NAND_GPBC_MAP11_ADR, (page >> 16) & 0xff);
	}
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank_sel | NAND_GPBC_MAP11_CMD, NAND_CMD_READSTART);
	irq_mask = NAND_GPBC_NI__INTACT | NAND_GPBC_NI__ERR;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irq_mask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irq_mask);
#endif
	if (NAND_GPBC_NI__ERR & result) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	}

	calc = 0xff;
	for (sect = 0; sect < chip->ecc.steps; sect++) {
		sect_base = (chip->ecc.size + chip->ecc.bytes) * sect;
		if (sect != chip->ecc.steps - 1) {
			ecc_base = sect_base + chip->ecc.size;
		} else {
			ecc_base = sect_base + gpbc->sectSize2 + gpbc->oobSkipSize + gpbc->sectSize3;
		}

		start = roundup(ecc_base - sizeof(u32) + 1, sizeof(u32));
		end = roundup(ecc_base + chip->ecc.bytes, sizeof(u32));

		for (; start < end; start += sizeof(u32)) {
			nand_gpbc_write32(NAND_GPBC_MAP00 | bank_sel | start, gpbc->memBase);
			*(u32 *)(gpbc->pPageBuf + start) = nand_gpbc_read32(gpbc->memBase + 0x10);
		}

		for (cnt = 0; cnt < chip->ecc.bytes; cnt++) {
			calc &= *(gpbc->pPageBuf + ecc_base + cnt);
		}
	}

	if (calc != 0xff) {
		return 0;
	}

	return 1;
}

static int nand_gpbc_isErasedPage(struct nand_gpbc_info *gpbc, int page, u8 *buf)
{
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	struct nand_chip *chip = &gpbc->chip;
	u32 page_offset = page & ((1 << (chip->phys_erase_shift - chip->page_shift)) - 1);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	/* compare DATA */
	if (!nand_gpbc_isErasedPageData(gpbc, buf)) {
		return 0;
	}

	/* compare ECC */
	if (!nand_gpbc_isErasedPageEcc(gpbc, page, buf)) {
		return 0;
	}

#ifdef CONFIG_MTD_NAND_MN2WS_BB
	set_bit_erased_map(page_offset, chip->erased_map);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */
	return 1;
}

static void nand_gpbc_restoreErasedData(struct nand_gpbc_info *gpbc, u8 *pBuf)
{
	struct nand_chip	*chip = &gpbc->chip;
	u32	sect;
	u32 correct_type = nand_gpbc_read32(gpbc->regBase + NAND_GPBC_NECCC);

	if (correct_type == 8) {
		//It is tested in only 8bit Ecc.
		for (sect = 0; sect < chip->ecc.steps; sect++) {
			*(u32 *)(pBuf + chip->ecc.size * sect + 0x2c0) = 0xffffffff;
		}
	}
}

static int nand_gpbc_readPageAllEcc(struct nand_gpbc_info *gpbc, uint8_t *pBuf, int page)
{
	struct mtd_info		*mtd = gpbc->mtd;
	struct nand_chip	*chip = &gpbc->chip;
	u32	irqMask;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
	u32	result, result2 = 0;
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	u32	result;
	u32	cnt;
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	int	bank;
	u32	bankSel;
	u32	bankPage;
	int	max_bitflips = 0;
	int	status;
	u32	dataPos;
	u32	oobPos;
	u32	sect;
	u32	pagePos;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
	struct nand_gpbc_cmd_desc_t *cmd_desc = gpbc->cmd_desc;
	u32 cmd_desc_phys = gpbc->cmd_desc_phys;
	u32 flash_page;
	u16 cmd_flags;
	u64 mem_ptr;
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */

	bank = gpbc->flash_bank;
	bankSel = bank << NAND_GPBC_BANK_SEL;
	bankPage = page & chip->pagemask;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	bankPage = bbt_translateBb(mtd, bankPage, bank);
	page = bankPage + (bank << (chip->chip_shift - chip->page_shift));
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	nand_gpbc_write32(NAND_GPBC_NTS__TS, gpbc->regBase + NAND_GPBC_NTS);	//spare
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
	flash_page = (NAND_GPBC_MAP10 | bankSel | bankPage);
	mem_ptr = (u64)gpbc->pageBufPhys;
	cmd_flags = (NAND_GPBC_CMD_FLAGS_TRANS_MAIN_SPARE |
		     NAND_GPBC_CMD_FLAGS_INT |
		     NAND_GPBC_CMD_FLAGS_BURST_LEN);
	nand_gpbc_set_cmd_desc_read_page(cmd_desc, flash_page,
					 cmd_flags, mem_ptr);
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP10 | bankSel | bankPage, NAND_GPBC_MAP10_MAIN_SPARE);	//transfer mode
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP10 | bankSel | bankPage, NAND_GPBC_MAP10_MAIN_SPARE);	//transfer mode
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	nand_gpbc_clearIntStatus(gpbc, bank);

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
	vaddr_purge_cache((u32)gpbc->pPageBuf, mtd->writesize + mtd->oobsize,
			  PURGE_CACHE_D_INV_BEFORE_DMA);
	nand_gpbc_write32(NAND_GPBC_NDME__FLAG, gpbc->regBase + NAND_GPBC_NDME);	//DMA ENABLE
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(NAND_GPBC_NECCE__ECCEN, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
	nand_gpbc_write32(NAND_GPBC_MAP10, gpbc->memBase);
	nand_gpbc_write32(NAND_GPBC_MAP10_BEAT0_CMD_DMA, gpbc->memBase + 0x10);	//BEAT0
	nand_gpbc_write32(cmd_desc_phys, gpbc->memBase + 0x10);	//BEAT1
	nand_gpbc_write32(0, gpbc->memBase + 0x10);	//BEAT2
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */
	nand_gpbc_write32(NAND_GPBC_MAP10 | bankSel | bankPage, gpbc->memBase);	//DMA ADDR
	nand_gpbc_write32(NAND_GPBC_MAP10_BEAT0_READ, gpbc->memBase + 0x10);	//BEAT0
	nand_gpbc_write32(gpbc->pageBufPhys, gpbc->memBase + 0x10);	//BEAT1
	nand_gpbc_write32(0, gpbc->memBase + 0x10);	//BEAT2
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */

	irqMask = NAND_GPBC_NI__DMA_CMD_COMP | NAND_GPBC_NI__TOUT;
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP10 | bankSel | bankPage, 0x2001);	//read-ahead
	irqMask = NAND_GPBC_NI__LDCMP | NAND_GPBC_NI__ERR;
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irqMask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irqMask);
#endif
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
	result2 = nand_gpbc_poll_cmd_dma(gpbc);
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NDME);	//DMA DISABLE
	vaddr_purge_cache((u32)gpbc->pPageBuf, mtd->writesize + mtd->oobsize,
			  PURGE_CACHE_D_INV_AFTER_DMA);
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	if (NAND_GPBC_NI__ERR & result) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	}
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
	if (result2 & NAND_GPBC_CDESC_STAT_ERR) {
		pr_err("\n## ERROR %s %s %d: result2 = 0x%08X ##\n",
		       __FILE__, __func__, __LINE__, result2);
		BUG();
	}
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(NAND_GPBC_NECCE__ECCEN, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_write32(NAND_GPBC_MAP01 | bankSel | bankPage, gpbc->memBase);
	for (cnt = 0; cnt < mtd->writesize + mtd->oobsize; cnt += 4) {
		*(u32 *)(gpbc->pPageBuf + cnt) = nand_gpbc_read32(gpbc->memBase + 0x10);
	}
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */

	pagePos = 0;
	dataPos = 0;
	oobPos = gpbc->oobSkipSize;
	for (sect = 0; sect < gpbc->chip.ecc.steps - 1; sect++) {
		memcpy(pBuf + dataPos, gpbc->pPageBuf + pagePos, gpbc->chip.ecc.size);
		pagePos += gpbc->chip.ecc.size;
		dataPos += gpbc->chip.ecc.size;
		memcpy(gpbc->chip.oob_poi + oobPos, gpbc->pPageBuf + pagePos, gpbc->chip.ecc.bytes);
		pagePos += gpbc->chip.ecc.bytes;
		oobPos += gpbc->chip.ecc.bytes;
	}
	memcpy(pBuf + dataPos, gpbc->pPageBuf + pagePos, gpbc->sectSize2);
	pagePos += gpbc->sectSize2;
	dataPos += gpbc->sectSize2;
	memcpy(gpbc->chip.oob_poi, gpbc->pPageBuf + pagePos, gpbc->oobSkipSize);
	pagePos += gpbc->oobSkipSize;
	memcpy(pBuf + dataPos, gpbc->pPageBuf + pagePos, gpbc->sectSize3);
	pagePos += gpbc->sectSize3;
	memcpy(gpbc->chip.oob_poi + oobPos, gpbc->pPageBuf + pagePos, mtd->oobsize - gpbc->oobSkipSize - gpbc->sectSize3);

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
	result2 &= NAND_GPBC_CDESC_STAT_FAIL;
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */
	result2 |= (result & NAND_GPBC_NI__ECCUNCOR);
	if (result2) {
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	result = nand_gpbc_checkIrq(gpbc, bank);
	if (NAND_GPBC_NI__ECCUNCOR & result) {
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
		status = nand_gpbc_isErasedPage(gpbc, page, pBuf);
		if (status) {
			nand_gpbc_restoreErasedData(gpbc, pBuf);
		} else {
			mtd->ecc_stats.failed++;
			pr_warn("\n\n## Warning ## %s %s %d: ECCUNCOR ##\n\n", __FILE__, __func__, __LINE__);
		}
	} else {
		max_bitflips = nand_gpbc_get_max_bitflips(gpbc, bank);
		if (max_bitflips) {
			nand_gpbc_update_correct_info(mtd, page, max_bitflips);
		}
		if (mtd->bitflip_threshold <= max_bitflips) {
			mtd->ecc_stats.corrected++;
		}
	}

	return max_bitflips;
}

static u32 nand_gpbc_readPageAllRaw(struct nand_gpbc_info *gpbc,
				    uint8_t *pBuf, int page)
{
	struct mtd_info		*mtd = gpbc->mtd;
	struct nand_chip *chip = &gpbc->chip;
	u32	irqMask;
	u32	result;
	u32	cnt;
	int	bank;
	u32	bankSel;
	u32	dataPos;
	u32	oobPos;
	u32	sect;
	u32	pagePos;

	bank = gpbc->flash_bank;
	bankSel = bank << NAND_GPBC_BANK_SEL;
	page = page & chip->pagemask;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	page = bbt_translateBb(mtd, page, bank);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	nand_gpbc_clearIntStatus(gpbc, bank);
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_CMD, NAND_CMD_READ0);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, 0);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, 0);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, page & 0xff);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, (page >> 8) & 0xff);
	if ((CONFIG_MTD_NAND_MN2WS_CHIP_CA_BYTES == 2) &&
	    (CONFIG_MTD_NAND_MN2WS_CHIP_PA_BYTES == 3)) {
		nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, (page >> 16) & 0xff);
	}
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_CMD, NAND_CMD_READSTART);
	irqMask = NAND_GPBC_NI__INTACT | NAND_GPBC_NI__ERR;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irqMask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irqMask);
#endif
	if (NAND_GPBC_NI__ERR & result) {
		pr_warn("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	}
	for (cnt = 0; cnt < mtd->writesize + mtd->oobsize; cnt += 4) {
		nand_gpbc_write32(NAND_GPBC_MAP00 | bankSel | cnt, gpbc->memBase);
		*(u32 *)(gpbc->pPageBuf + cnt) = nand_gpbc_read32(gpbc->memBase + 0x10);
	}

	pagePos = 0;
	dataPos = 0;
	oobPos = gpbc->oobSkipSize;
	for (sect = 0; sect < gpbc->chip.ecc.steps - 1; sect++) {
		memcpy(pBuf + dataPos, gpbc->pPageBuf + pagePos, gpbc->chip.ecc.size);
		pagePos += gpbc->chip.ecc.size;
		dataPos += gpbc->chip.ecc.size;
		memcpy(gpbc->chip.oob_poi + oobPos, gpbc->pPageBuf + pagePos, gpbc->chip.ecc.bytes);
		pagePos += gpbc->chip.ecc.bytes;
		oobPos += gpbc->chip.ecc.bytes;
	}
	memcpy(pBuf + dataPos, gpbc->pPageBuf + pagePos, gpbc->sectSize2);
	pagePos += gpbc->sectSize2;
	dataPos += gpbc->sectSize2;
	memcpy(gpbc->chip.oob_poi, gpbc->pPageBuf + pagePos, gpbc->oobSkipSize);
	pagePos += gpbc->oobSkipSize;
	memcpy(pBuf + dataPos, gpbc->pPageBuf + pagePos, gpbc->sectSize3);
	pagePos += gpbc->sectSize3;
	memcpy(gpbc->chip.oob_poi + oobPos, gpbc->pPageBuf + pagePos, mtd->oobsize - gpbc->oobSkipSize - gpbc->sectSize3);

	return 0;
}

static int nand_gpbc_writePageAllEcc(struct nand_gpbc_info *gpbc,
				     const uint8_t *pBuf, int page)
{
	struct mtd_info		*mtd = gpbc->mtd;
	struct nand_chip	*chip = &gpbc->chip;
	u32	dataPos;
	u32	oobPos;
	u32	sect;
	u32	pagePos;
	u32	irqMask;
	u32 result;
	int	bank;
	u32	bankSel;
	u32	bankPage;

	bank = gpbc->flash_bank;
	bankSel = bank << NAND_GPBC_BANK_SEL;
	bankPage = page & chip->pagemask;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	bankPage = bbt_translateBb(mtd, bankPage, bank);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	pagePos = 0;
	dataPos = 0;
	oobPos = gpbc->oobSkipSize;
	for (sect = 0; sect < gpbc->chip.ecc.steps - 1; sect++) {
		memcpy(gpbc->pPageBuf + pagePos, pBuf + dataPos, gpbc->chip.ecc.size);
		pagePos += gpbc->chip.ecc.size;
		dataPos += gpbc->chip.ecc.size;
		memcpy(gpbc->pPageBuf + pagePos, gpbc->chip.oob_poi + oobPos, gpbc->chip.ecc.bytes);
		pagePos += gpbc->chip.ecc.bytes;
		oobPos += gpbc->chip.ecc.bytes;
	}
	memcpy(gpbc->pPageBuf + pagePos, pBuf + dataPos, gpbc->sectSize2);
	pagePos += gpbc->sectSize2;
	dataPos += gpbc->sectSize2;
	memcpy(gpbc->pPageBuf + pagePos, gpbc->chip.oob_poi, gpbc->oobSkipSize);
	pagePos += gpbc->oobSkipSize;
	memcpy(gpbc->pPageBuf + pagePos, pBuf + dataPos, gpbc->sectSize3);
	pagePos += gpbc->sectSize3;
	memcpy(gpbc->pPageBuf + pagePos, gpbc->chip.oob_poi + oobPos, mtd->oobsize - gpbc->oobSkipSize - gpbc->sectSize3);

	nand_gpbc_write32(NAND_GPBC_NTS__TS, gpbc->regBase + NAND_GPBC_NTS);	//spare
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP10 | bankSel | bankPage, NAND_GPBC_MAP10_MAIN_SPARE);	//transfer mode
	nand_gpbc_clearIntStatus(gpbc, bank);
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(NAND_GPBC_NECCE__ECCEN, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
	vaddr_purge_cache((u32)gpbc->pPageBuf, mtd->writesize + mtd->oobsize, PURGE_CACHE_D_PURGE);	//L1/L2 cache clean
	nand_gpbc_write32(NAND_GPBC_NDME__FLAG, gpbc->regBase + NAND_GPBC_NDME);	//enable DMA
	nand_gpbc_write32(NAND_GPBC_MAP10 | bankSel | bankPage, gpbc->memBase);	//DMA ADDR
	nand_gpbc_write32(NAND_GPBC_MAP10_BEAT0_WRITE, gpbc->memBase + 0x10);	//BEAT0
	nand_gpbc_write32(gpbc->pageBufPhys, gpbc->memBase + 0x10);	//BEAT1
	nand_gpbc_write32(0, gpbc->memBase + 0x10);	//BEAT2

	irqMask = NAND_GPBC_NI__DMA_CMD_COMP | NAND_GPBC_NI__TOUT;
#else /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	nand_gpbc_write32(NAND_GPBC_MAP01 | bankSel | bankPage, gpbc->memBase);
	for (pagePos = 0; pagePos < mtd->writesize + mtd->oobsize; pagePos += 4) {
		nand_gpbc_write32(*(u32 *)(gpbc->pPageBuf + pagePos), gpbc->memBase + 0x10);
	}
	irqMask = NAND_GPBC_NI__PROCMP | NAND_GPBC_NI__ERR;
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irqMask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irqMask);
#endif
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NDME);	//disable DMA
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_DMA */
	if (NAND_GPBC_NI__ERR & ~NAND_GPBC_NI__PROFAIL & result) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	}

	gpbc->status = (result & NAND_GPBC_NI__PROFAIL) ? NAND_STATUS_FAIL : 0;

	return 0;
}

static int nand_gpbc_writePageAllRaw(struct nand_gpbc_info *gpbc,
				      const uint8_t *pBuf, int page)
{
	struct mtd_info		*mtd = gpbc->mtd;
	struct nand_chip	*chip = &gpbc->chip;
	u32	dataPos;
	u32	oobPos;
	u32	sect;
	u32	pagePos;
	u32	irqMask;
	u32 result;
	int	bank;
	u32	bankSel;
	u32	bankPage;
	u32	cnt;

	bank = gpbc->flash_bank;
	bankSel = bank << NAND_GPBC_BANK_SEL;
	bankPage = page & chip->pagemask;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	bankPage = bbt_translateBb(mtd, bankPage, bank);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	pagePos = 0;
	dataPos = 0;
	oobPos = gpbc->oobSkipSize;
	for (sect = 0; sect < gpbc->chip.ecc.steps - 1; sect++) {
		memcpy(gpbc->pPageBuf + pagePos, pBuf + dataPos, gpbc->chip.ecc.size);
		pagePos += gpbc->chip.ecc.size;
		dataPos += gpbc->chip.ecc.size;
		memcpy(gpbc->pPageBuf + pagePos, gpbc->chip.oob_poi + oobPos, gpbc->chip.ecc.bytes);
		pagePos += gpbc->chip.ecc.bytes;
		oobPos += gpbc->chip.ecc.bytes;
	}
	memcpy(gpbc->pPageBuf + pagePos, pBuf + dataPos, gpbc->sectSize2);
	pagePos += gpbc->sectSize2;
	dataPos += gpbc->sectSize2;
	memcpy(gpbc->pPageBuf + pagePos, gpbc->chip.oob_poi, gpbc->oobSkipSize);
	pagePos += gpbc->oobSkipSize;
	memcpy(gpbc->pPageBuf + pagePos, pBuf + dataPos, gpbc->sectSize3);
	pagePos += gpbc->sectSize3;
	memcpy(gpbc->pPageBuf + pagePos, gpbc->chip.oob_poi + oobPos, mtd->oobsize - gpbc->oobSkipSize - gpbc->sectSize3);

	nand_gpbc_clearIntStatus(gpbc, bank);
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_CMD, NAND_CMD_SEQIN);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, 0);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, 0);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, page & 0xff);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, (page >> 8) & 0xff);
	if ((CONFIG_MTD_NAND_MN2WS_CHIP_CA_BYTES == 2) &&
	    (CONFIG_MTD_NAND_MN2WS_CHIP_PA_BYTES == 3)) {
		nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_ADR, (page >> 16) & 0xff);
	}
	for (cnt = 0; cnt < mtd->writesize + mtd->oobsize; cnt += 4) {
		nand_gpbc_write32(NAND_GPBC_MAP00 | bankSel | cnt, gpbc->memBase);
		nand_gpbc_write32(*(u32 *)(gpbc->pPageBuf + cnt), gpbc->memBase + 0x10);
	}
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bankSel | NAND_GPBC_MAP11_CMD, NAND_CMD_PAGEPROG);
	irqMask = NAND_GPBC_NI__INTACT | NAND_GPBC_NI__ERR;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irqMask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irqMask);
#endif
	if (NAND_GPBC_NI__ERR & ~NAND_GPBC_NI__PROFAIL & result) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	}

	gpbc->status = (result & NAND_GPBC_NI__PROFAIL) ? NAND_STATUS_FAIL : 0;

	return 0;
}

static int nand_gpbc_readPageOob(struct nand_gpbc_info *gpbc, int page)
{
	struct mtd_info		*mtd = gpbc->mtd;
	struct nand_chip	*chip = &gpbc->chip;
	u32	irqMask;
	u32	result;
	u32	cnt;
	u8	*pBuf;
	int	bank;
	u32	bankSel;
	u32	bankPage;

	bank = gpbc->flash_bank;
	bankSel = bank << NAND_GPBC_BANK_SEL;
	bankPage = page & chip->pagemask;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	bankPage = bbt_translateBb(mtd, bankPage, bank);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	nand_gpbc_write32(NAND_GPBC_NTS__TS, gpbc->regBase + NAND_GPBC_NTS);	//spare
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP10 | bankSel | bankPage, NAND_GPBC_MAP10_SPARE);	//transfer mode
	nand_gpbc_clearIntStatus(gpbc, bank);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP10 | bankSel | bankPage, 0x2001);	//read-ahead
	irqMask = NAND_GPBC_NI__LDCMP | NAND_GPBC_NI__ERR;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irqMask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irqMask);
#endif
	if (NAND_GPBC_NI__ERR & result) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	}
	nand_gpbc_write32(NAND_GPBC_MAP01 | bankSel | bankPage, gpbc->memBase);
	pBuf = chip->oob_poi;
	for (cnt = 0; cnt < mtd->oobsize; cnt += 4) {
		*(u32 *)(pBuf + cnt) = nand_gpbc_read32(gpbc->memBase + 0x10);
	}

	return 0;
}

static int nand_gpbc_writePageOob(struct nand_gpbc_info *gpbc, int page)
{
	struct mtd_info		*mtd = gpbc->mtd;
	struct nand_chip	*chip = &gpbc->chip;
	u32	irqMask;
	u32	result;
	u32	cnt;
	int	bank;
	u32	bankSel;
	u32	bankPage;

	bank = gpbc->flash_bank;
	bankSel = bank << NAND_GPBC_BANK_SEL;
	bankPage = page & chip->pagemask;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	bankPage = bbt_translateBb(mtd, bankPage, bank);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	nand_gpbc_write32(NAND_GPBC_NTS__TS, gpbc->regBase + NAND_GPBC_NTS);	//spare
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP10 | bankSel | bankPage, NAND_GPBC_MAP10_SPARE);	//transfer mode
	nand_gpbc_clearIntStatus(gpbc, bank);
	nand_gpbc_write32(NAND_GPBC_MAP01 | bankSel | bankPage, gpbc->memBase);
	for (cnt = 0; cnt < mtd->oobsize; cnt += 4) {
		nand_gpbc_write32(*(u32 *)(gpbc->chip.oob_poi + cnt), gpbc->memBase + 0x10);
	}
	irqMask = NAND_GPBC_NI__PROCMP | NAND_GPBC_NI__ERR;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irqMask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irqMask);
#endif
	if (NAND_GPBC_NI__ERR & ~NAND_GPBC_NI__PROFAIL & result) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	} else if (NAND_GPBC_NI__PROFAIL & result) {
		return -EIO;
	}

	return 0;
}

//----------------------------
//-------- ecc func ----------
//----------------------------
static void nand_gpbc_hwctl(struct mtd_info *mtd, int mode)
{
	pr_err("nand_gpbc_hwctl called unexpectedly\n");
	BUG();
}

static int nand_gpbc_calculate(struct mtd_info *mtd, const uint8_t *dat,
			uint8_t *ecc_code)
{
	pr_err("nand_gpbc_calculate called unexpectedly\n");
	BUG();
	return -EIO;
}

static int nand_gpbc_correct(struct mtd_info *mtd, uint8_t *dat, uint8_t *read_ecc,
		      uint8_t *calc_ecc)
{
	pr_err("nand_gpbc_correct called unexpectedly\n");
	BUG();
	return -EIO;
}

static int nand_gpbc_read_page_raw(struct mtd_info *mtd, struct nand_chip *chip,
				   uint8_t *buf, int oob_required, int page)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	return nand_gpbc_readPageAllRaw(gpbc, buf, page);
}

static int nand_gpbc_write_page_raw(struct mtd_info *mtd, struct nand_chip *chip,
				     const uint8_t *buf, int oob_required)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	return nand_gpbc_writePageAllRaw(gpbc, buf, gpbc->page);
}

static int nand_gpbc_read_page(struct mtd_info *mtd, struct nand_chip *chip,
			       uint8_t *buf, int oob_required, int page)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	return nand_gpbc_readPageAllEcc(gpbc, buf, page);
}

static int nand_gpbc_read_subpage(struct mtd_info *mtd, struct nand_chip *chip,
			    uint32_t offs, uint32_t len, uint8_t *buf)
{
	pr_err("nand_gpbc_read_subpage called unexpectedly\n");
	BUG();
	return 0;
}

static int nand_gpbc_write_page(struct mtd_info *mtd, struct nand_chip *chip,
				const uint8_t *buf, int oob_required)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	return nand_gpbc_writePageAllEcc(gpbc, buf, gpbc->page);
}

static int nand_gpbc_read_oob(struct mtd_info *mtd, struct nand_chip *chip,
			      int page)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	return nand_gpbc_readPageOob(gpbc, page);
}

static int nand_gpbc_write_oob(struct mtd_info *mtd, struct nand_chip *chip, int page)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	return nand_gpbc_writePageOob(gpbc, page);
}


//----------------------------
//------- chip func ----------
//----------------------------
static uint8_t nand_gpbc_read_byte(struct mtd_info *mtd)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	uint8_t result = 0xff;

	if (gpbc->buf.head < gpbc->buf.tail) {
		result = gpbc->buf.buf[gpbc->buf.head++];
	}
	return result;
}

static u16 nand_gpbc_read_word(struct mtd_info *mtd)
{
	pr_err("nand_gpbc_read_word called unexpectedly\n");
	BUG();
	return 0;
}

static void nand_gpbc_write_buf(struct mtd_info *mtd, const uint8_t *buf, int len)
{
	pr_err("nand_gpbc_write_buf called unexpectedly\n");
	BUG();
}

static void nand_gpbc_read_buf(struct mtd_info *mtd, uint8_t *buf, int len)
{
	pr_err("nand_gpbc_read_buf called unexpectedly\n");
	BUG();
}

static void nand_gpbc_select_chip(struct mtd_info *mtd, int chip)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	gpbc->flash_bank = chip;
}

static void nand_gpbc_cmd_ctrl(struct mtd_info *mtd, int dat, unsigned int ctrl)
{
	pr_err("nand_gpbc_cmd_ctrl called unexpectedly\n");
	BUG();
}

/* when use auto detect, don't set init_size. */
#if 0
static int nand_gpbc_init_size(struct mtd_info *mtd, struct nand_chip *this, u8 *id_data)
{
	int busw;

	this->cellinfo = id_data[2];
	mtd->writesize = CONFIG_MTD_NAND_MN2WS_CHIP_PAGESIZE;
	mtd->oobsize = CONFIG_MTD_NAND_MN2WS_CHIP_OOBSIZE;
	mtd->erasesize = CONFIG_MTD_NAND_MN2WS_CHIP_BLOCKSIZE;

	busw = 0; /* 8bit only */

	return busw;
}
#endif

static void nand_gpbc_cmdfunc(struct mtd_info *mtd, unsigned int cmd, int col,
			      int page)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);

	switch (cmd) {
	case NAND_CMD_PAGEPROG:
		break;
	case NAND_CMD_STATUS:
		nand_gpbc_read_status(gpbc);
		break;
	case NAND_CMD_READID:
	case NAND_CMD_PARAM:
		nand_gpbc_reset_buf(gpbc);
		/*sometimes ManufactureId read from register is not right
		 * e.g. some of Micron MT29F32G08QAA MLC NAND chips
		 * So here we send READID cmd to NAND insteand
		 * */
		nand_read_id(gpbc, gpbc->aId, MTD_NAND_MN2WS_ID_READ_NUM);
		break;
	case NAND_CMD_READ0:
	case NAND_CMD_SEQIN:
		gpbc->page = page;
		break;
	case NAND_CMD_RESET:
		nand_gpbc_resetBank(gpbc, gpbc->flash_bank);
		break;
	case NAND_CMD_READOOB:
		/* TODO: Read OOB data */
		break;
	default:
		pr_err(": unsupported command received 0x%x\n", cmd);
		break;
	}
}

static int nand_gpbc_waitfunc(struct mtd_info *mtd, struct nand_chip *this)
{
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	int status = gpbc->status;

	gpbc->status = 0;
	return status;
}

static void nand_gpbc_erase_cmd(struct mtd_info *mtd, int page)
{
	struct nand_chip *chip = mtd->priv;
	struct nand_gpbc_info *gpbc = mtd_to_gpbc(mtd);
	u32 irqMask;
	u32 result;
	int bank;
	u32 bankSel;
	u32 bankPage;

	bank = gpbc->flash_bank;
	bankSel = bank << NAND_GPBC_BANK_SEL;
	bankPage = page & chip->pagemask;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	bankPage = bbt_translateBb(mtd, bankPage, bank);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	nand_gpbc_clearIntStatus(gpbc, bank);
#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP10 | bankSel | bankPage, NAND_GPBC_MAP10_ERASE);
	irqMask = NAND_GPBC_NI__ERACMP | NAND_GPBC_NI__ERR;
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	result = nand_gpbc_waitIrq(gpbc, bank, irqMask);
#else
	result = nand_gpbc_pollIrq(gpbc, bank, irqMask);
#endif
	if (NAND_GPBC_NI__ERR & ~NAND_GPBC_NI__ERAFAIL & result) {
		pr_err("\n## ERROR %s %s %d: result = 0x%08X ##\n", __FILE__, __func__, __LINE__, result);
		BUG();
	}

	gpbc->status = (result & NAND_GPBC_NI__ERAFAIL) ? NAND_STATUS_FAIL : 0;
}


//----------------------------
//-------- init func ---------
//----------------------------
static int nand_gpbc_initHw1(struct nand_gpbc_info *gpbc)
{
	u32		temp32;

	temp32 = nand_gpbc_read32(gpbc->regBase + NAND_GPBC_NECCE);
	if (!temp32) {
		pr_err("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		BUG();
	}

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_RESET
	/* Reset GPBC */
	nand_gpbc_mask_write32(0, NAND_GPBC_UNIT_RESET, (void*)NAND_GPBC_UNIT_RESET_BASE);
	nand_gpbc_mask_write32(NAND_GPBC_UNIT_RESET, NAND_GPBC_UNIT_RESET, (void*)NAND_GPBC_UNIT_RESET_BASE);
#endif
	
	/* disable WP */
	nand_gpbc_write32(NAND_GPBC_NWPD__WPD, gpbc->regBase + NAND_GPBC_NWPD);

	//Enable R/B pin of all banks
	nand_gpbc_write32(NAND_GPBC_NRBE__MASK, gpbc->regBase + NAND_GPBC_NRBE);

	/* ecc size */
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CORRECT_ERROR_BITS,
			  gpbc->regBase + NAND_GPBC_NECCC);

	//pulse
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CHIP_NCDWR, gpbc->regBase + NAND_GPBC_NCDWR);
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CHIP_NCAAD, gpbc->regBase + NAND_GPBC_NCAAD);
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CHIP_NRW, gpbc->regBase + NAND_GPBC_NRW);
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CHIP_NRDCP, gpbc->regBase + NAND_GPBC_NRDCP);
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CHIP_NRWLP, gpbc->regBase + NAND_GPBC_NRWLP);
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CHIP_NRWHP, gpbc->regBase + NAND_GPBC_NRWHP);
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CHIP_NCES, gpbc->regBase + NAND_GPBC_NCES);
	nand_gpbc_write32(CONFIG_MTD_NAND_MN2WS_CHIP_NRR, gpbc->regBase + NAND_GPBC_NRR);

	return 0;
}

static void nand_read_id(struct nand_gpbc_info *gpbc, u8 *nand_id, int num)
{
	u32 bank = gpbc->flash_bank << NAND_GPBC_BANK_SEL;
	int i = 0;
	uint32_t id;
	static unsigned long checked_bank;

#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NECCE);	//ecc
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank | NAND_GPBC_MAP11_CMD, NAND_CMD_READID);
	nand_gpbc_index32(gpbc, NAND_GPBC_MAP11 | bank | NAND_GPBC_MAP11_ADR, 0);

	for (i = 0; i < num; i++) {
		nand_gpbc_write32(NAND_GPBC_MAP11 | bank | NAND_GPBC_MAP11_DAT, gpbc->memBase);
		id = nand_gpbc_read32(gpbc->memBase + 0x10);
		nand_id[i] =  (u8)(id & 0xff);
		nand_gpbc_write_byte_to_buf(gpbc, (uint8_t)id);
	}

	if (!__test_and_set_bit(gpbc->flash_bank, &checked_bank)) {
		pr_info("nand-gpbc: device ID (bank%d):", gpbc->flash_bank);
		for (i = 0; i < num; i++)
			pr_cont(" 0x%02x", nand_id[i]);
		pr_cont("\n");
	}
}

static void nand_gpbc_initHw2(struct nand_gpbc_info *gpbc)
{
	struct mtd_info *mtd = gpbc->mtd;
	struct nand_chip *chip = &gpbc->chip;
	u32		temp32;

	//
	// If the chip doesn't support ONFI,
	//   set default values of registers classified into AUTOCONF reg.

	//Set SPARE_AREA_SKIP_BYTES
	nand_gpbc_write32(NAND_GPBC_OOB_SKIP_SIZE, gpbc->regBase + NAND_GPBC_NSASK);

	//Set SPARE_AREA_MARKER
	nand_gpbc_write32(0xffff, gpbc->regBase + NAND_GPBC_NSAMK);

	//Set PAGES_PER_BLOCK
	temp32 = mtd->erasesize / mtd->writesize;
	nand_gpbc_write32(temp32, gpbc->regBase + NAND_GPBC_NBPNUM);

	//Set DEVICE_MAIN_AREA_SIZE
	nand_gpbc_write32(mtd->writesize, gpbc->regBase + NAND_GPBC_NMASZ);

	//Set DEVICE_SPARE_AREA_SIZE
	nand_gpbc_write32(mtd->oobsize, gpbc->regBase + NAND_GPBC_NSASZ);

#if defined(CONFIG_PROC_MN2WS0300) || defined(CONFIG_PROC_MN2WS0310) || defined(CONFIG_PROC_MN2WS0320)
	//Set CFG_NUM_DATA_BLOCKS
	nand_gpbc_write32(NAND_GPBC_ECCNUM,
			  gpbc->regBase + NAND_GPBC_NDB);

	//Set CHIP_INTERLEAVE_ENABLE
	nand_gpbc_write32(NAND_GPBC_NCIEAIR__AIR,
			  gpbc->regBase + NAND_GPBC_NCIEAIR);
#endif /* CONFIG_PROC_MN2WS0300 || CONFIG_PROC_MN2WS0310 || CONFIG_PROC_MN2WS0320 */

	//Set MANUFACTURE_ID
	nand_gpbc_write32(gpbc->aId[0], gpbc->regBase + NAND_GPBC_NMID);

	//Set DEVICE_WIDTH
	if (!(chip->options & NAND_BUSWIDTH_16)) {
		nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NDW);
	} else {
		pr_err("\n## ERROR %s %s %d: supported organization is only \"x8\" ##\n",
		       __FILE__, __func__, __LINE__);
		BUG();
	}

	//Set DEVICE_NUMBER
	nand_gpbc_write32(0x00000001, gpbc->regBase + NAND_GPBC_NDNUM);

	//Set Address cycle
	if (CONFIG_MTD_NAND_MN2WS_CHIP_CA_BYTES == 2 && CONFIG_MTD_NAND_MN2WS_CHIP_PA_BYTES == 2) {
		nand_gpbc_write32(NAND_GPBC_NTRADC__TRADC, gpbc->regBase + NAND_GPBC_NTRADC);
	}
}


#ifdef CONFIG_MTD_NAND_MN2WS_BB

#if defined(NAND_MN2WS_AREA_BLOCKS_BOOT) && defined(NAND_MN2WS_AREA_BLOCKS_ALT)
/* nothing to do */
#elif defined(NAND_GPBC_UNIT_BTBLOCKS) && defined(NAND_GPBC_UNIT_MTBLOCKS)
#define NAND_MN2WS_BLOCK_SIZE_OLD_COMPAT
#else /* NAND_MN2WS_AREA_BLOCKS_BOOT && NAND_MN2WS_AREA_BLOCKS_ALT */
#error Error: Could not find area size settings.
#endif /* NAND_MN2WS_AREA_BLOCKS_BOOT && NAND_MN2WS_AREA_BLOCKS_ALT */

/* AREA_ORDER {BOOT_ORDER, DATA_ORDER, ALT_ORDER, BBM_ORDER} */
#define NAND_MN2WS_AREA_ORDER_LASTBLOCK \
	{NAND_AREA_ID_BOOT, NAND_AREA_ID_DATA, NAND_AREA_ID_ALT, NAND_AREA_ID_BBM}
#define NAND_MN2WS_AREA_ORDER_FIRSTBLOCK \
	{NAND_AREA_ID_BBM, NAND_AREA_ID_ALT, NAND_AREA_ID_DATA, NAND_AREA_ID_BOOT}
#ifdef NAND_MN2WS_AREA_ORDER
#if (NAND_MN2WS_AREA_ORDER == NAND_MN2WS_AREA_ORDER_BDAM)
#define NAND_MN2WS_AREA_ORDER_FLEXIBLE \
	{NAND_AREA_ID_BOOT, NAND_AREA_ID_DATA, NAND_AREA_ID_ALT, NAND_AREA_ID_BBM}
#elif (NAND_MN2WS_AREA_ORDER == NAND_MN2WS_AREA_ORDER_BDMA)
#define NAND_MN2WS_AREA_ORDER_FLEXIBLE \
	{NAND_AREA_ID_BOOT, NAND_AREA_ID_DATA, NAND_AREA_ID_BBM, NAND_AREA_ID_ALT}
#elif (NAND_MN2WS_AREA_ORDER == NAND_MN2WS_AREA_ORDER_BADM)
#define NAND_MN2WS_AREA_ORDER_FLEXIBLE \
	{NAND_AREA_ID_BOOT, NAND_AREA_ID_ALT, NAND_AREA_ID_DATA, NAND_AREA_ID_BBM}
#elif (NAND_MN2WS_AREA_ORDER == NAND_MN2WS_AREA_ORDER_BAMD)
#define NAND_MN2WS_AREA_ORDER_FLEXIBLE \
	{NAND_AREA_ID_BOOT, NAND_AREA_ID_ALT, NAND_AREA_ID_BBM, NAND_AREA_ID_DATA}
#elif (NAND_MN2WS_AREA_ORDER == NAND_MN2WS_AREA_ORDER_BMDA)
#define NAND_MN2WS_AREA_ORDER_FLEXIBLE \
	{NAND_AREA_ID_BOOT, NAND_AREA_ID_BBM, NAND_AREA_ID_DATA, NAND_AREA_ID_ALT}
#elif (NAND_MN2WS_AREA_ORDER == NAND_MN2WS_AREA_ORDER_BMAD)
#define NAND_MN2WS_AREA_ORDER_FLEXIBLE \
	{NAND_AREA_ID_BOOT, NAND_AREA_ID_BBM, NAND_AREA_ID_ALT, NAND_AREA_ID_DATA}
#else
#error Error: Invalid  NAND_MN2WS_AREA_ORDER.
#endif
#else /* NAND_MN2WS_AREA_ORDER */
/* dummy: not used NAND_MN2WS_AREA_ORDER_FLEXIBLE */
#define NAND_MN2WS_AREA_ORDER_FLEXIBLE NAND_MN2WS_AREA_ORDER_LASTBLOCK
#endif /* NAND_MN2WS_AREA_ORDER */

#define SET_AREA_ORDER(dst, order)		\
do {						\
	typeof(dst) __temp = order;		\
	memcpy(dst, __temp, sizeof(dst));	\
} while (0)

static int nand_gpbc_setAreaInfo(struct nand_chip *chip,
				 struct nand_bbt_area_info *area_info,
				 u32 areaBlocks[][NAND_MAX_CHIPS], u32 chip_num)
{
	u32 area, block, size, i;
	u32 chipBlocks = 1 << (chip->chip_shift - chip->bbt_erase_shift);
	u32 order[NAND_AREA_ID_NUM];

	if ((chip->bbt_td->options ^ chip->bbt_md->options) &
	    (NAND_BBT_FLEXIBLE | NAND_BBT_LASTBLOCK)) {
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		return -EINVAL;
	}

	if (chip->bbt_td->options & NAND_BBT_FLEXIBLE) {
		SET_AREA_ORDER(order, NAND_MN2WS_AREA_ORDER_FLEXIBLE);
	} else if (chip->bbt_td->options & NAND_BBT_LASTBLOCK) {
		SET_AREA_ORDER(order, NAND_MN2WS_AREA_ORDER_LASTBLOCK);
	} else {
		SET_AREA_ORDER(order, NAND_MN2WS_AREA_ORDER_FIRSTBLOCK);
	}

	block = chipBlocks * chip_num;
	for (i = 0; i < NAND_AREA_ID_NUM; i++) {
		area = order[i];
		size = areaBlocks[area][chip_num];
		area_info[area].startBlock = block;
		area_info[area].blockNum = size;
		block += size;
	}

	return 0;
}

static int nand_gpbc_allocBbmResources(struct nand_gpbc_info *gpbc)
{
	u32		cnt;
	u32		chipBlocks = 1 << (gpbc->chip.chip_shift - gpbc->chip.bbt_erase_shift);
	/* {{BOOT}, {DATA}, {ALT}, {BBM}} */
	u32 areaBlocks[NAND_AREA_ID_NUM][NAND_MAX_CHIPS] =
#ifdef NAND_MN2WS_BLOCK_SIZE_OLD_COMPAT
		{NAND_GPBC_UNIT_BTBLOCKS, {0}, NAND_GPBC_UNIT_MTBLOCKS, {0}};
#else /* NAND_MN2WS_BLOCK_SIZE_OLD_COMPAT */
	{NAND_MN2WS_AREA_BLOCKS_BOOT, {0}, NAND_MN2WS_AREA_BLOCKS_ALT, {0}};
#endif /* NAND_MN2WS_BLOCK_SIZE_OLD_COMPAT */

	for (cnt = 0; cnt < gpbc->chip.numchips; cnt++) {
#ifdef NAND_MN2WS_BLOCK_SIZE_OLD_COMPAT
		areaBlocks[NAND_AREA_ID_ALT][cnt] -= NAND_BBM_BLOCK_NUM;
#endif /* NAND_MN2WS_BLOCK_SIZE_OLD_COMPAT */
		areaBlocks[NAND_AREA_ID_BBM][cnt] = NAND_BBM_BLOCK_NUM;
		areaBlocks[NAND_AREA_ID_DATA][cnt] = chipBlocks -
			(areaBlocks[NAND_AREA_ID_BOOT][cnt] + areaBlocks[NAND_AREA_ID_ALT][cnt] +
			 areaBlocks[NAND_AREA_ID_BBM][cnt]);
	}

	gpbc->chip.pBbmBuf = NULL;
	gpbc->chip.pBlockBuf = NULL;
	gpbc->chip.psBbm = nand_gpbc_ABbm;
	for (cnt = 0; cnt < NAND_MAX_CHIPS; cnt++) {
		struct nand_bbm		*psBbm;

		psBbm = &gpbc->chip.psBbm[cnt];
		psBbm->mtBlocks = areaBlocks[NAND_AREA_ID_ALT][cnt] +
			areaBlocks[NAND_AREA_ID_BBM][cnt];
		psBbm->bbmPages = -1;
		psBbm->bbMapSize = -1;
		psBbm->bbListSize = -1;
		psBbm->usedMtBlocks = -1;
		psBbm->altBlocks = -1;
		psBbm->pBbMap = NULL;
		psBbm->psBbList = NULL;
		psBbm->area = NULL;
	}

	for (cnt = 0; cnt < gpbc->chip.numchips; cnt++) {
		struct nand_bbm		*psBbm;

		psBbm = &gpbc->chip.psBbm[cnt];
		psBbm->bbMapSize = ((chipBlocks + 8 * 4 - 1) / (8 * 4)) * 4;
		psBbm->bbListSize = sizeof(struct nand_bbList) * psBbm->mtBlocks;
		if (gpbc->chip.bbt_td->options & NAND_BBT_NO_OOB) {
			psBbm->bbmPages = (psBbm->bbMapSize + psBbm->bbListSize + 4 * 2
					   + sizeof(gpbc->chip.bbt_td->pages[0])
					   + sizeof(gpbc->chip.bbt_md->pages[0])
					   + (1 << gpbc->chip.page_shift)
					   + sizeof(gpbc->chip.bbt_td->len)
					   + roundup(sizeof(gpbc->chip.bbt_td->version[0]), sizeof(u32)) - 1)
				/ (1 << gpbc->chip.page_shift);
		} else {
			psBbm->bbmPages = (psBbm->bbMapSize + psBbm->bbListSize + 4 * 2
					   + sizeof(gpbc->chip.bbt_td->pages[0])
					   + sizeof(gpbc->chip.bbt_md->pages[0])
					   + (1 << gpbc->chip.page_shift) - 1)
				/ (1 << gpbc->chip.page_shift);
		}
		psBbm->pBbMap = kmalloc(psBbm->bbMapSize, GFP_KERNEL);
		if (!psBbm->pBbMap) {
			return -ENOMEM;
		}
		psBbm->psBbList = kmalloc(psBbm->bbListSize, GFP_KERNEL);
		if (!psBbm->psBbList) {
			return -ENOMEM;
		}
		psBbm->area = vmalloc(sizeof(struct nand_bbt_area_info) * NAND_AREA_ID_NUM);
		if (!psBbm->area) {
			return -ENOMEM;
		}
		if (nand_gpbc_setAreaInfo(&gpbc->chip, psBbm->area, areaBlocks, cnt)) {
			return -EINVAL;
		}
	}

	gpbc->chip.pBbmBuf = vmalloc(gpbc->chip.psBbm[0].bbmPages * (gpbc->mtd->writesize + gpbc->mtd->oobsize));
	if (!gpbc->chip.pBbmBuf) {
		return -ENOMEM;
	}

	return 0;
}

static void nand_gpbc_freeBbmResources(struct nand_gpbc_info *gpbc)
{
	u32		cnt;

	if (gpbc->chip.pBbmBuf) {
		vfree(gpbc->chip.pBbmBuf);
	}
	for (cnt = 0; cnt < gpbc->chip.numchips; cnt++) {
		struct nand_bbm		*psBbm;

		psBbm = &gpbc->chip.psBbm[cnt];
		if (psBbm->pBbMap) {
			kfree(psBbm->pBbMap);
		}
		if (psBbm->psBbList) {
			kfree(psBbm->psBbList);
		}
	}
}
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

#ifdef CONFIG_MTD_NAND_MN2WS_BB
static int nand_gpbc_alloc_resources(struct nand_gpbc_info *gpbc)
{
	int pages_per_block = 1 << (gpbc->chip.bbt_erase_shift - gpbc->chip.page_shift);
	size_t size;
	int ret = 0;

	ret = nand_gpbc_allocBbmResources(gpbc);
	if (ret) {
		return ret;
	}

	size = pages_per_block * (gpbc->mtd->writesize + gpbc->mtd->oobsize);
	gpbc->chip.pBlockBuf = vmalloc(size);
	if (!gpbc->chip.pBlockBuf) {
		return -ENOMEM;
	}

	size = DIV_ROUND_UP(pages_per_block, 32) * sizeof(u32);
	gpbc->chip.erased_map = vmalloc(size);
	if (!gpbc->chip.erased_map) {
		return -ENOMEM;
	}
	clear_erased_map(&gpbc->chip);

	return 0;
}

static void nand_gpbc_free_resources(struct nand_gpbc_info *gpbc)
{
	if (gpbc->chip.erased_map) {
		vfree(gpbc->chip.erased_map);
	}

	if (gpbc->chip.pBlockBuf) {
		vfree(gpbc->chip.pBlockBuf);
	}

	nand_gpbc_freeBbmResources(gpbc);
}
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

static void nand_gpbc_initInfo1(struct nand_gpbc_info *gpbc)
{
	gpbc->nist[0] = NAND_GPBC_NIST0;
	gpbc->nist[1] = NAND_GPBC_NIST1;
	gpbc->nie[0] = NAND_GPBC_NIE0;
	gpbc->nie[1] = NAND_GPBC_NIE1;
}

int board_nand_init_gpbc(struct mtd_info *mtd)
{
	struct nand_chip	*chip;
	int err = 0;
	struct nand_gpbc_info *gpbc;
#ifdef CONFIG_MTD_CMDLINE_PARTS
	int num_parts;
#endif  /* CONFIG_MTD_CMDLINE_PARTS */
	int		status;

	gpbc = kzalloc(sizeof(struct nand_gpbc_info), GFP_KERNEL);
	if (!gpbc) {
		err = -ENOMEM;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out;
	}

	gpbc->mtd = mtd;
	chip = &gpbc->chip;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	chip->flags = 0;
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	gpbc->regBase = (void *)NAND_GPBC_UNIT_REG_BASE;
	if (!gpbc->regBase) {
		err = -ENOMEM;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_err;
	}
	gpbc->memBase = (void *)NAND_GPBC_UNIT_MEM_BASE;
	if (!gpbc->memBase) {
		err = -ENOMEM;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_err;
	}

	nand_gpbc_initInfo1(gpbc);

	status = nand_gpbc_initHw1(gpbc);
	if (status) {
		err = status;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_err2;
	}

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	gpbc->irq = NAND_GPBC_IRQ_ID;
	gpbc->event = 0;
	init_waitqueue_head(&gpbc->queue);
	nand_gpbc_write32(0, gpbc->regBase + gpbc->nie[0]);
	nand_gpbc_write32(0, gpbc->regBase + gpbc->nie[1]);
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NDMIE);
	nand_gpbc_write32(NAND_GPBC_NGINTE__INTEN, gpbc->regBase + NAND_GPBC_NGINTE);
	mn_intc_clear(gpbc->irq);
	status = request_irq(gpbc->irq, nand_gpbc_irq, 0, "nand-gpbc", gpbc);
	if (status) {
		err = status;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_irq;
	}
#endif

	mtd->priv = &gpbc->chip;
	mtd->bitflip_threshold = NAND_GPBC_BITFLIP_THRESHOLD;

	chip->priv = gpbc;
	chip->IO_ADDR_R = 0;
	chip->IO_ADDR_W = 0;

	chip->read_byte = nand_gpbc_read_byte;
	chip->read_word = nand_gpbc_read_word;
	chip->write_buf = nand_gpbc_write_buf;
	chip->read_buf = nand_gpbc_read_buf;
	chip->select_chip = nand_gpbc_select_chip;


	chip->cmd_ctrl = nand_gpbc_cmd_ctrl;
	/* when use auto detect, don't set init_size. */
	/* chip->init_size = nand_gpbc_init_size; */

	chip->cmdfunc = nand_gpbc_cmdfunc;
	chip->waitfunc = nand_gpbc_waitfunc;



	chip->update_correct_info = nand_gpbc_update_correct_info;
	chip->get_correct_info = nand_gpbc_get_correct_info;
	chip->clear_correct_info = nand_gpbc_clear_correct_info;

	chip->bbt_options = NAND_GPBC_BBT_OPTIONS;
	chip->options = NAND_GPBC_OPTIONS;

	chip->chip_delay = 0;	//unused param

	status = nand_scan_ident(gpbc->mtd, NAND_GPBC_FLASH_BANKS, NULL);
	if (status) {
		err = status;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_err2;
	}

	/* overwrite oobsize by user setting */
	mtd->oobsize = CONFIG_MTD_NAND_MN2WS_CHIP_OOBSIZE;

	/* sanity check for auto-detected configuration */
	BUG_ON(mtd->writesize != CONFIG_MTD_NAND_MN2WS_CHIP_PAGESIZE);
	BUG_ON(mtd->oobsize < CONFIG_MTD_NAND_MN2WS_CHIP_OOBSIZE);
	BUG_ON(mtd->erasesize != CONFIG_MTD_NAND_MN2WS_CHIP_BLOCKSIZE);

	/* overwrite nand_chip structures */
	chip->erase_cmd = nand_gpbc_erase_cmd;
	chip->badblockbits = 1; /* unused param */

#ifndef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
	gpbc->pPageBuf = kmalloc(mtd->writesize + mtd->oobsize, GFP_KERNEL);
	if (!gpbc->pPageBuf) {
		err = -ENOMEM;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_err2;
	}
#else
	if (NAND_GPBC_DMA_BURST < cache_line_size()) {
		if (cache_line_size() % NAND_GPBC_DMA_BURST) {
			BUG();
		}
		gpbc->dmaAlignSize = cache_line_size();
	} else {
		if (NAND_GPBC_DMA_BURST % cache_line_size()) {
			BUG();
		}
		gpbc->dmaAlignSize = NAND_GPBC_DMA_BURST;
	}
	gpbc->pPageBufRaw = kmalloc(ALIGN(mtd->writesize + mtd->oobsize, gpbc->dmaAlignSize) + gpbc->dmaAlignSize, GFP_KERNEL);
	if (!gpbc->pPageBufRaw) {
		err = -ENOMEM;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_err2;
	}

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ
	gpbc->ucarea = (struct nand_gpbc_ucarea_t *)CONFIG_NAND_UCAREA_ADDR;
	if (!gpbc->ucarea) {
		err = -ENOMEM;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_err2;
	}
	gpbc->ucarea_phys = (u32)gpbc->ucarea;

	gpbc->cmd_desc = &gpbc->ucarea->cmd_desc;
	gpbc->cmd_desc_phys = gpbc->ucarea_phys +
		offsetof(struct nand_gpbc_ucarea_t, cmd_desc);
#endif /* CONFIG_MTD_NAND_MN2WS_GPBC_CDMA_READ */
#endif

	nand_gpbc_initHw2(gpbc);

#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
	gpbc->pPageBuf = (void *)ALIGN((u32)gpbc->pPageBufRaw, gpbc->dmaAlignSize);
	gpbc->pageBufPhys = virt_to_phys(gpbc->pPageBuf);

	if ((PAGE_SIZE < cache_line_size()) ||
	    (PAGE_SIZE % cache_line_size()) ||
	    (mtd->writesize < cache_line_size()) ||
	    (mtd->writesize % cache_line_size()) ||
	    (PAGE_SIZE < mtd->writesize) ||
	    (PAGE_SIZE % mtd->writesize) ||
	    ((u32)gpbc->pPageBuf % cache_line_size())) {
		BUG();
	}
#endif

	//the chip spec has been reflected in "options" on nand_gpbc_getInfoFromId.
	//So, clear the options not supported by the controller.
	gpbc->chip.options &= ~(NAND_CACHEPRG);

	gpbc->pagesPerBlock = mtd->erasesize / mtd->writesize;
	gpbc->sectSize3 = NAND_GPBC_ECCBYTES * (gpbc->mtd->writesize / NAND_GPBC_ECCSIZE - 1);
	gpbc->sectSize2 = NAND_GPBC_ECCSIZE - gpbc->sectSize3;
	gpbc->oobSkipSize = NAND_GPBC_OOB_SKIP_SIZE;

	gpbc->chip.bbt_td = &nand_gpbc_BbtMasterDesc;
	gpbc->chip.bbt_md = &nand_gpbc_BbtMirrorDesc;
#ifdef CONFIG_MTD_NAND_MN2WS_BB
	gpbc->chip.bbt_flag = &nand_gpbc_BbtFlagDesc;
#endif /* CONFIG_MTD_NAND_MN2WS_BB */
	gpbc->chip.badblock_pattern = &nand_gpbc_BbPatternDesc;

	gpbc->chip.ecc.mode = NAND_ECC_HW_SYNDROME;
	gpbc->chip.ecclayout = &nand_gpbc_EccLayout;
	gpbc->chip.ecc.layout = gpbc->chip.ecclayout;
	gpbc->chip.ecc.bytes = NAND_GPBC_ECCBYTES;
	gpbc->chip.ecc.strength = CONFIG_MTD_NAND_MN2WS_CORRECT_ERROR_BITS;
	gpbc->chip.ecc.size = NAND_GPBC_ECCSIZE;
	gpbc->chip.ecc.prepad = 0;
	gpbc->chip.ecc.postpad = 0;

	gpbc->chip.ecc.hwctl = nand_gpbc_hwctl;
	gpbc->chip.ecc.calculate = nand_gpbc_calculate;
	gpbc->chip.ecc.correct = nand_gpbc_correct;
	gpbc->chip.ecc.read_page_raw = nand_gpbc_read_page_raw;
	gpbc->chip.ecc.write_page_raw = nand_gpbc_write_page_raw;
	gpbc->chip.ecc.read_page = nand_gpbc_read_page;
	gpbc->chip.ecc.read_subpage = nand_gpbc_read_subpage;
	gpbc->chip.ecc.write_page = nand_gpbc_write_page;
	gpbc->chip.ecc.write_oob_raw = nand_gpbc_write_oob;
	gpbc->chip.ecc.read_oob_raw = nand_gpbc_read_oob;
	gpbc->chip.ecc.read_oob = nand_gpbc_read_oob;
	gpbc->chip.ecc.write_oob = nand_gpbc_write_oob;

#ifdef CONFIG_MTD_NAND_MN2WS_BB
	err = nand_gpbc_alloc_resources(gpbc);
	if (err) {
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_bbm;
	}
#endif /* CONFIG_MTD_NAND_MN2WS_BB */

	err = nand_scan_tail(gpbc->mtd);
	if (err) {
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_bbt;
	}

	if (!(chip->options & NAND_SKIP_BBTSCAN)) {
		err = chip->scan_bbt(mtd);
		if (err) {
			printk("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
			goto out_bbt;
		}
	}

#ifdef CONFIG_MTD_PARTITIONS
#ifdef CONFIG_MTD_CMDLINE_PARTS
	num_parts = parse_mtd_partitions(gpbc->mtd,
					 data->chip.part_probe_types,
					 &gpbc->parts, 0);
	if (num_parts > 0) {
		err = add_mtd_partitions(gpbc->mtd, gpbc->parts, num_parts);
		if (err) {
			pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
			goto out_part;
		}
	} else {
		err = -EINVAL;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_part;
	}
#else /* CONFIG_MTD_CMDLINE_PARTS */
	/* platform specific function to set partitions */
	if (data->chip.partitions) {
		gpbc->parts = data->chip.partitions;
		err = add_mtd_partitions(gpbc->mtd, gpbc->parts,
					 data->chip.nr_partitions);
		if (err) {
			pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
			goto out_part;
		}
	}
#endif /* CONFIG_MTD_CMDLINE_PARTS */
#endif /* CONFIG_MTD_PARTITIONS */

	status = add_mtd_device(gpbc->mtd);
	if (status) {
		err = -1;
		pr_warn("\n## ERROR %s %s %d ##\n", __FILE__, __func__, __LINE__);
		goto out_mtd;
	}
	goto out;

 out_mtd:
#ifdef CONFIG_MTD_PARTITIONS
	/* Deregister partitions */
	del_mtd_partitions(mtd);
 out_part:
#ifdef CONFIG_MTD_CMDLINE_PARTS
	/* freeing partition info if necessary */
	if (gpbc->parts && gpbc->parts != data->chip.partitions) {
		kfree(gpbc->parts);
	}
#endif //CONFIG_MTD_CMDLINE_PARTS
#endif	//CONFIG_MTD_PARTITIONS
 out_bbt:
	if (chip->bbt) {
		kfree(chip->bbt);
	}
#ifdef CONFIG_MTD_NAND_MN2WS_BB
 out_bbm:
	nand_gpbc_free_resources(gpbc);
#endif /* CONFIG_MTD_NAND_MN2WS_BB */
 out_err2:
#ifndef CONFIG_MTD_NAND_MN2WS_GPBC_DMA
	if (gpbc->pPageBuf) {
		kfree(gpbc->pPageBuf);
	}
#else
	if (gpbc->pPageBufRaw) {
		kfree(gpbc->pPageBufRaw);
	}
#endif
#ifdef CONFIG_MTD_NAND_MN2WS_GPBC_IRQ
	free_irq(gpbc->irq, gpbc);
 out_irq:
	nand_gpbc_write32(0, gpbc->regBase + NAND_GPBC_NGINTE);
#endif
 out_err:
	if (chip->buffers) {
		kfree(chip->buffers);
	}
	if (gpbc) {
		kfree(gpbc);
	}
 out:

	return err;
}
